@@ -1,13 +1,16 @@
 """Provide a package for python-openevse-http."""
 from __future__ import annotations
 
+import asyncio
 import datetime
 import logging
 from typing import Optional
 
+import aiohttp
 import requests  # type: ignore
 
 from .const import MAX_AMPS, MIN_AMPS
+from .exceptions import AuthenticationError, ParseJSONError, HTTPError
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -27,17 +30,175 @@ states = {
     255: "disabled",
 }
 
+ERROR_AUTH_FAILURE = "Authorization failure"
+ERROR_TOO_MANY_RETRIES = "Too many retries"
+ERROR_UNKNOWN = "Unknown"
 
-class AuthenticationError(Exception):
-    """Exception for authentication errors."""
+MAX_FAILED_ATTEMPTS = 5
+
+SIGNAL_CONNECTION_STATE = "websocket_state"
+STATE_CONNECTED = "connected"
+STATE_DISCONNECTED = "disconnected"
+STATE_STARTING = "starting"
+STATE_STOPPED = "stopped"
 
 
-class ParseJSONError(Exception):
-    """Exception for JSON parsing errors."""
+class OpenEVSEWebsocket:
+    """Represent a websocket connection to a OpenEVSE charger."""
 
+    def __init__(
+        self,
+        server,
+        callback,
+        user=None,
+        password=None,
+    ):
+        """Initialize a OpenEVSEWebsocket instance.
+        Parameters:
+            server (openevse address):
+                A connected instance.
+            callback (Runnable):
+                Called when interesting events occur. Provides arguments:
+                   msgtype (str): Message type or SIGNAL_* constant
+                   data (str): Current STATE_* or websocket payload contents
+                   error (str): Error message if any or None
+        """
+        self.session = aiohttp.ClientSession()
+        self.uri = self._get_uri(server)
+        self._user = (user,)
+        self._password = (password,)
+        self.callback = callback
+        self.subscriptions = ["message"]
+        self._state = None
+        self.failed_attempts = 0
+        self._error_reason = None
 
-class HTTPError(Exception):
-    """Exception for HTTP errors."""
+    @property
+    def state(self):
+        """Return the current state."""
+        return self._state
+
+    @state.setter
+    def state(self, value):
+        """Set the state."""
+        self._state = value
+        _LOGGER.debug("Websocket %s", value)
+        self.callback(SIGNAL_CONNECTION_STATE, value, self._error_reason)
+        self._error_reason = None
+
+    @staticmethod
+    def _get_uri(server):
+        """Generate the websocket URI."""
+        return server._url("/:/ws").replace("http", "ws")
+
+    async def running(self):
+        """Open a persistent websocket connection and act on events."""
+        self.state = STATE_STARTING
+
+        try:
+            if self._user and self._password:
+                async with self.session.ws_connect(
+                    self.uri,
+                    heartbeat=15,
+                    auth=aiohttp.BasicAuth(self._user, self._password),
+                ) as ws_client:
+                    self.state = STATE_CONNECTED
+                    self.failed_attempts = 0
+
+                    async for message in ws_client:
+                        if self.state == STATE_STOPPED:
+                            break
+
+                        if message.type == aiohttp.WSMsgType.TEXT:
+                            msg = message.json()
+                            msgtype = msg["type"]
+
+                            if msgtype not in self.subscriptions:
+                                _LOGGER.debug("Ignoring: %s", msg)
+                                continue
+
+                            else:
+                                self.callback(msgtype, msg, None)
+
+                        elif message.type == aiohttp.WSMsgType.CLOSED:
+                            _LOGGER.warning("AIOHTTP websocket connection closed")
+                            break
+
+                        elif message.type == aiohttp.WSMsgType.ERROR:
+                            _LOGGER.error("AIOHTTP websocket error")
+                            break
+            else:
+                async with self.session.ws_connect(
+                    self.uri,
+                    heartbeat=15,
+                ) as ws_client:
+                    self.state = STATE_CONNECTED
+                    self.failed_attempts = 0
+
+                    async for message in ws_client:
+                        if self.state == STATE_STOPPED:
+                            break
+
+                        if message.type == aiohttp.WSMsgType.TEXT:
+                            msg = message.json()
+                            msgtype = msg["type"]
+
+                            if msgtype not in self.subscriptions:
+                                _LOGGER.debug("Ignoring: %s", msg)
+                                continue
+
+                            else:
+                                self.callback(msgtype, msg, None)
+
+                        elif message.type == aiohttp.WSMsgType.CLOSED:
+                            _LOGGER.warning("AIOHTTP websocket connection closed")
+                            break
+
+                        elif message.type == aiohttp.WSMsgType.ERROR:
+                            _LOGGER.error("AIOHTTP websocket error")
+                            break
+
+        except aiohttp.ClientResponseError as error:
+            if error.code == 401:
+                _LOGGER.error("Credentials rejected: %s", error)
+                self._error_reason = ERROR_AUTH_FAILURE
+            else:
+                _LOGGER.error("Unexpected response received: %s", error)
+                self._error_reason = ERROR_UNKNOWN
+            self.state = STATE_STOPPED
+        except (aiohttp.ClientConnectionError, asyncio.TimeoutError) as error:
+            if self.failed_attempts >= MAX_FAILED_ATTEMPTS:
+                self._error_reason = ERROR_TOO_MANY_RETRIES
+                self.state = STATE_STOPPED
+            elif self.state != STATE_STOPPED:
+                retry_delay = min(2 ** (self.failed_attempts - 1) * 30, 300)
+                self.failed_attempts += 1
+                _LOGGER.error(
+                    "Websocket connection failed, retrying in %ds: %s",
+                    retry_delay,
+                    error,
+                )
+                self.state = STATE_DISCONNECTED
+                await asyncio.sleep(retry_delay)
+        except Exception as error:  # pylint: disable=broad-except
+            if self.state != STATE_STOPPED:
+                _LOGGER.exception("Unexpected exception occurred: %s", error)
+                self._error_reason = ERROR_UNKNOWN
+                self.state = STATE_STOPPED
+        else:
+            if self.state != STATE_STOPPED:
+                self.state = STATE_DISCONNECTED
+                await asyncio.sleep(5)
+
+    async def listen(self):
+        """Close the listening websocket."""
+        self.failed_attempts = 0
+        while self.state != STATE_STOPPED:
+            await self.running()
+
+    def close(self):
+        """Close the listening websocket."""
+        self.state = STATE_STOPPED
 
 
 class OpenEVSE:
